! Copyright (C) 2017-2024 Matthieu Ancellin
! See LICENSE file at <https://github.com/capytaine/libDelhommeau>

module floating_point_precision

  implicit none

  integer, parameter :: pre = kind(1e0)

end module floating_point_precision
